/* wc - print the number of bytes, words, and lines in files
   Copyright (C) 85, 91, 95, 1996 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* Written by Paul Rubin, phr@ocf.berkeley.edu
   and David MacKenzie, djm@gnu.ai.mit.edu. */
/* Modified by P.Syrowatka, syro@mycity.at;
   Adding code for file-mapping in the case of lseek-able input file */


#include <config.h>

#include <stdio.h>
#include <getopt.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "system.h"
#include "error.h"

/* Size of atomic reads. */
#define BUFFER_SIZE (16 * 1024)

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

#ifndef MAP_FAILED
#define MAP_FAILED -1
#endif

#define FILE_MAP  1
#define FIFO_MAP  2

#define PAGE_SIZE   (getpagesize())

#define MAP_INIT_SIZE  (PAGE_SIZE * 4096)

int safe_read ();

/* The name this program was run with. */
char *program_name;

/* Cumulative number of lines, words, and chars in all files so far. */
static unsigned long total_lines, total_words, total_chars;

/* Which counts to print. */
static int print_lines, print_words, print_chars;

/* Nonzero if we have ever read the standard input. */
static int have_read_stdin;

/* The error code to return to the system. */
static int exit_status;

/* If nonzero, display usage information and exit.  */
static int show_help;

/* If nonzero, print the version on standard output then exits.  */
static int show_version;

/* Werden im Signal-Handler gebraucht => global */
static unsigned long chars, stat_base = 0;
static off_t map_off;
short fifo_cont = 1;
int globfd = 0, i;

static struct option const longopts[] =
{
  {"bytes", no_argument, NULL, 'c'},
  {"chars", no_argument, NULL, 'c'},
  {"lines", no_argument, NULL, 'l'},
  {"words", no_argument, NULL, 'w'},
  {"help", no_argument, &show_help, 1},
  {"version", no_argument, &show_version, 1},
  {NULL, 0, NULL, 0}
};

static void
usage (int status)
{
  if (status != 0)
    fprintf (stderr, _("Try `%s --help' for more information.\n"),
	     program_name);
  else
    {
      printf (_("\
Usage: %s [OPTION]... [FILE]...\n\
"),
	      program_name);
      printf (_("\
Print line, word, and byte counts for each FILE, and a total line if\n\
more than one FILE is specified.  With no FILE, or when FILE is -,\n\
read standard input.\n\
  -l, --lines            print the newline counts\n\
  -w, --words            print the word counts\n\
  -c, --bytes, --chars   print the byte counts\n\
      --help             display this help and exit\n\
      --version          output version information and exit\n\
"));
    }
  exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
}

static void
write_counts (long unsigned int lines, long unsigned int words,
	      long unsigned int chars, const char *file)
{
  if (print_lines)
    printf ("%7lu", lines);
  if (print_words)
    {
      if (print_lines)
	putchar (' ');
      printf ("%7lu", words);
    }
  if (print_chars)
    {
      if (print_lines || print_words)
	putchar (' ');
      printf ("%7lu", chars);
    }
  if (*file)
    printf (" %s", file);
  putchar ('\n');
}


/* Set to SIGPIPE */

void sigbus(int sig) {

  printf("got SIGBUS : chars %d\n",chars);
  exit (-6);
}

void sigh(int sig) {
  struct pstat_struct pstat;

  pstat.addr = stat_base + i;
  /*
    printf("SIGPIPE, ask bufsize from 0x%lx (chars %d, i %d)\n",pstat.addr,chars,i); */
  if (ioctl(globfd,PM_GETSTAT,&pstat) == -1) {
    perror("sigh");
    exit (-5);
  }
  if (! pstat.sigflag) {
    printf("SIGPIPE : SIGPIPE not to this pipe ..\n");
    exit (-8); 
  }
  chars = pstat.tot_rbuf + i;
  /*
  printf("\tReaders         : %d\n",pstat.readers);
  printf("\tWriters         : %d\n",pstat.writers);
  printf("\tBuffer size     : %lu  (0x%lx)\n",pstat.bsize,pstat.bsize);
  printf("\tRead alligned   : %lu  (0x%lx)\n",pstat.sec_rbuf,pstat.sec_rbuf);
  printf("\tRead total      : %lu  (0x%lx)\n",pstat.tot_rbuf,pstat.tot_rbuf);
  printf("\tAuto-commit     : %lu  (0x%lx)\n",pstat.updsize,pstat.updsize);
  printf("\tVMA-start       : %lu  (0x%lx)\n",pstat.vmstart,pstat.vmstart);
  printf("\tI/O-Position    : %lu  (0x%lx)\n",pstat.io_pos,pstat.io_pos);
  printf("chars : %d\n",chars);
getchar();
  */
  fifo_cont = 0;
  signal(SIGPIPE, sigh);
  signal(SIGBUS, sigbus);
  return;
}



char * next_mmap(char * unmap, unsigned long len, 
		 int prot, int flag, 
		 int fd, off_t off) {
  char * retv;
  /*  
  printf("do next_mmap() : unmap %lu , offset %lu...\n",len,off);
  */
  munmap(unmap, len);
  if ((retv = mmap(NULL,len,prot,flag,fd,off)) == (char *)MAP_FAILED) {
    perror("next_mmap");
    exit (-2);
  }
  /*
  printf("next_mmap() : return 0x%lx\n",retv);
  */
  return retv;
}
static void
wc (int fd, const char *file)
{
  register int bytes_read;
  register int in_word = 0;
  register unsigned long lines, words;
  int   do_it_traditional = TRUE,
        map_mode, csum = 0;

  off_t current_pos = 0, end_pos = 0;
  struct stat stats;
  struct pstat_struct pstat;

  globfd = fd;

  lines = words = chars = 0;

  /* When counting only bytes, save some line- and word-counting
     overhead.  If FD is a `regular' Unix file, using lseek is enough
     to get its `size' in bytes.  Otherwise, read blocks of BUFFER_SIZE
     bytes at a time until EOF.  Note that the `size' (number of bytes)
     that wc reports is smaller than stats.st_size when the file is not
     positioned at its beginning.  That's why the lseek calls below are
     necessary.  For example the command
     `(dd ibs=99k skip=1 count=0; ./wc -c) < /etc/group'
     should make wc report `0' bytes.  */

  /* Zuerst muss festgestellt werden ob die Datei "uberhaupt map-bar 
   * ist. Das wird mit dem lseek()-Test gemacht; dieser liefert dann auch
   * gleich die f"urs Mappen ben"otigte gr"osse der Datei. (In st_mode steht
   * mehr als nur die Protection !!) */

  map_mode = FILE_MAP;

  /* Ist folgender Test positiv, kann gemapped werden .. */
  if (fstat (fd, &stats) == 0 && ((S_ISREG (stats.st_mode)
      && (current_pos = lseek (fd, (off_t) 0, SEEK_CUR)) != -1
      && (end_pos = lseek (fd, (off_t) 0, SEEK_END)) != -1)) || 
      (S_ISFIFO(stats.st_mode) && (map_mode = FIFO_MAP)))
    { /* Mapping - Zweig. */
      off_t diff;
      int map_flag = MAP_SHARED, map_prot = PROT_READ;
      do_it_traditional = FALSE;
      /* Be careful here.  The current position may actually be
	 beyond the end of the file.  As in the example above.  */
      if (map_mode == FILE_MAP)
	chars = (diff = end_pos - current_pos) < 0 ? 0 : diff;
      else {
	/*
	printf ("[%d] Map_Mode : FIFO_MAP - fd : %d\n",getpid(),fd);
	*/
	if (ioctl(fd,PM_BUFOFF,&map_off) == -1) {
	  do_it_traditional = TRUE;
	  perror("ioctl(PM_BUFOFF)");
	  goto stdio;
	}
	if (signal(SIGPIPE, sigh) == SIG_ERR){
	  do_it_traditional = TRUE;
	  perror("signal");
	  goto stdio;
	}
	chars = MAP_INIT_SIZE;
	map_flag |= MAP_RDSTREAM;
      }
      if (chars > 0 && (print_words || print_lines || map_mode == FIFO_MAP)) 
	/* Nur wenn die Datei groesser 0 ist und mehr als Byte-Count 
	 * verlangt ist, wird gemapped. */  
	{
	  char *buf = NULL; /* Map-Pointer */
	  unsigned long mchars = 0;
	  int dlevel = 1;
	  if ((buf = mmap (NULL, chars, map_prot, map_flag, fd, map_off))\
	      != (char *) MAP_FAILED) { /* gemapptes File counten ... */
	    if ((map_mode == FIFO_MAP) && (ioctl(fd,PM_DLEV,&dlevel) == -1)) {
	      do_it_traditional = TRUE;
	      perror("ioctl (PM_DLEV)");
	    }
            do {
	    stat_base = (unsigned long) buf;
	    if (!print_words) /* Worte werden nicht gezaehlt ... */
	      {
	      for (i = 0; i < chars; i++)
		if (buf[i] == '\n') ++lines; 
	      }
	    else       /* Count lines, words and bytes ... */
	      {
	      for (i = 0; i < chars; i++)
		switch (buf[i])
		  {
		  case '\n':
		    lines++;
		    /* Fall through. */
		  case '\r':
		  case '\f':
		  case '\t':
		  case '\v':
		  case ' ':
		    if (in_word)
		      {
			in_word = 0;
			words++;
		      }
		    break;
		  default:
		    in_word = 1;
		    break;
		  }
	      }

	    /* Wurden char-Bytes abgearbeitet, muss bei einer FIFO nocheinmal 
	     * eingeblendet werden. War das Pipe-Ende schon signalisiert, 
	     * wurde im Signalhandler das fifo_cont Flag schon geloescht. */

	    if ((map_mode == FIFO_MAP) && fifo_cont) {
	      mchars += chars; 
	      buf = next_mmap(buf, chars, map_prot, map_flag, 
			      fd, map_off + mchars); 
	    }
	    }
	    while ((map_mode == FIFO_MAP) && fifo_cont);
	    munmap(buf, chars);
	    if (map_mode == FIFO_MAP) chars += mchars;
	    
	  }
          else  do_it_traditional = TRUE; /* Wenn mapping nicht geht ...*/
	}
           lseek (fd, (off_t) current_pos, SEEK_SET);   /* R"ucksetzen ... */
    }
  stdio :
  if (do_it_traditional)   /* Datei ist nicht map-bar; alles wie bisher */
    {
    char buf[BUFFER_SIZE + 1];
    if (print_chars && !print_words && !print_lines)
      { /* Hier ergibt sich der Unterschied, dass der Fall wo durch
	 * lseek() die Byte-Anzahl ermittelt werden kann, schon im
	 * mmap-test enthalten ist. Daher kann hier sofort mit dem 
	 * Lesen begonnen werden. */
	while ((bytes_read = safe_read (fd, buf, BUFFER_SIZE)) > 0)
	  {
	    chars += bytes_read;
	  }
	if (bytes_read < 0)
	  {
	    error (0, errno, "%s", file);
	    exit_status = 1;
	  }
      }
    else if (!print_words)
      {
	/* Use a separate loop when counting only lines or lines and bytes --
	   but not words.  */
	while ((bytes_read = safe_read (fd, buf, BUFFER_SIZE)) > 0)
	  {
	    register char *p = buf;
	    
	    while ((p = memchr (p, '\n', (buf + bytes_read) - p)))
	      {
		++p;
		++lines;
	      }
	    chars += bytes_read;
	  }
	if (bytes_read < 0)
	  {
	    error (0, errno, "%s", file);
	    exit_status = 1;
	  }
      }
    else     /* Count lines, words and bytes ... */
      {
	while ((bytes_read = safe_read (fd, buf, BUFFER_SIZE)) > 0)
	  {
	    register char *p = buf;
	    
	    chars += bytes_read;
	    do
	      {
		switch (*p++)
		  {
		  case '\n':
		    lines++;
		    /* Fall through. */
		  case '\r':
		  case '\f':
		  case '\t':
		  case '\v':
		  case ' ':
		    if (in_word)
		      {
			in_word = 0;
			words++;
		      }
		    break;
		  default:
		    in_word = 1;
		    break;
		  }
	      }
	    while (--bytes_read);
	  } 
	if (bytes_read < 0)
	  {
	    error (0, errno, "%s", file);
	    exit_status = 1;
	  }
	if (in_word)
	  words++;
      }
    }
  write_counts (lines, words, chars, file);
  total_lines += lines;
  total_words += words;
  total_chars += chars;
}

static void
wc_file (const char *file)
{
  /* Ist in der Liste der Files ein "-", dann soll dazwischen wieder von 
     stdin gelesen werden */
  if (!strcmp (file, "-"))
    {
      have_read_stdin = 1;
      wc (0, file);
    }
  else
    {
      int fd = open (file, O_RDONLY);
      if (fd == -1)
	{
	  error (0, errno, "%s", file);
	  exit_status = 1;
	  return;
	}
      wc (fd, file);
      if (close (fd))
	{
	  error (0, errno, "%s", file);
	  exit_status = 1;
	}
    }
}

int
main (int argc, char **argv)
{
  int optc;
  int nfiles;

  program_name = argv[0];
  /* Fuer die Unterstuetzung diverser laenderspezifischer Einstellungen */
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  exit_status = 0;
  print_lines = print_words = print_chars = 0;
  total_lines = total_words = total_chars = 0;

  /* Einlesen und Auswerten der Parameter */

  while ((optc = getopt_long (argc, argv, "clw", longopts, (int *) 0)) != EOF)
    switch (optc)
      {
      case 0:
	break;

      case 'c':
	print_chars = 1;
	break;

      case 'l':
	print_lines = 1;
	break;

      case 'w':
	print_words = 1;
	break;

      default:
	usage (1);
      }

  /* Versionsinfo anzeigen */
  if (show_version)
    {
      printf ("mwc - %s %s\n", GNU_PACKAGE, VERSION);
      exit (EXIT_SUCCESS);
    }

  /* Usage/Hilfe Info ausgeben */
  if (show_help)
    usage (0);

  if (print_lines + print_words + print_chars == 0)
    print_lines = print_words = print_chars = 1;

  nfiles = argc - optind;

  /* Wenn keine Datei als Parameter angegeben ist, wird von stdin gelesen */
  if (nfiles == 0)
    {
      have_read_stdin = 1;
      wc (0, "");
    }
  else
    /* Ansonsten werden alle als Parameter angegebenen Files nacheinander
       bearbeitet. */
    {
      for (; optind < argc; ++optind)
	wc_file (argv[optind]);
      /* Wenns mehrere waren, dann wird eine Gesammtsumme ausgegeben. */
      if (nfiles > 1)
	write_counts (total_lines, total_words, total_chars, _("total"));
    }

  if (have_read_stdin && close (0))
    error (EXIT_FAILURE, errno, "-");

  exit (exit_status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
}








